(require (lib "list.ss"))
(require (lib "pretty.ss"))
(require (lib "string.ss"))

;; Graph operations:
;; The search function is a general search function.  Depending on the search-type input, it calls
;; different search methods.  These search methods are what you will be implementing in ps2.
;; The input search-type is a string used to define the type of search to be performed.
;; The input useExtendedList? is a boolean value that if true means you should use an extended list 
;;    to perform the search.
;; The input graph is of the graph represenation shown later in search.scm.
;; The input start is a node-id indicating which node in the graph the search should start at.
;; The input goal is a node-id indicating which node in the graph the search should end at.
(define (search search-type useExtendedList? graph start goal)
  (cond ((equal? search-type "BFS") 
         (bfs (get-nodes graph) 
              (get-edges graph) 
              useExtendedList? 
              '()
              (list (list start))
              start 
              goal))
        ((equal? search-type "DFS")
         (dfs (get-nodes graph)
              (get-edges graph)
              useExtendedList?
              '()
              (list (list start))
              start
              goal))
        ((equal? search-type "HC")
         (hill-climbing (get-nodes graph)
                        (get-edges graph)
                        useExtendedList?
                        '()
                        (list (list start))
                        start
                        goal))
        ((equal? search-type "Branch and Bound")
         (branch-and-bound (get-nodes graph)
                           (get-edges graph)
                           useExtendedList?
                           '()
                           (list (list start))
                           start
                           goal))
        ((equal? search-type "A*")
         (a-star (get-nodes graph)
                 (get-edges graph)
                 '()
                 (list (list start))
                 start
                 goal))
        (else "Search type has not yet been written")))

;; A helper function for finding a part of a rule based on the keyword.
;; Graphs are divided into nodes and edges.  This function is a helper function
;; that gets the list associated with each keyword, in this case, either 'NODES
;; or 'EDGES.  You won't need this function to use this function to complete ps2.scm.
(define (find-graph-part keyword default)
  (lambda (lst)
      (let ((p (filter (lambda (lst-part)
                 (member keyword lst-part)) lst))
            (chunk (member keyword (car lst))))
        (if p
            (cdar p)
            default))))

;; A helper function to get the last element in a list.
;; This function might be useful when trying to find the last node-id in a path.
(define (get-last lst)
  (car (last-pair lst)))

;; Node operations:
;; get-nodes returns a list of all the nodes in the graph.
(define get-nodes (find-graph-part 'NODES '()))
;; node-heuristic returns the heuristic value of a node.
(define (node-heuristic node)
  (cadr node))
;; node-id returns the node id value for a node.
(define (node-id node)
  (car node))
;; get-heuristic returns the heurstic value of a node in the graph
;; The input id is the node id for the node that you want to get the heuristic value of.
;; The input nodes is a list of all the nodes in a graph. 
(define (get-node id nodes)
  (assoc id nodes))

(define (get-heuristic id nodes)
  (let ([node (get-node id nodes)])
    (if node (node-heuristic node)
        (error (expr->string (list "Couldn't find a node with this id:" id))))))

;; - gets a list of all node id values connected to a given node.
(define (get-connected-nodes node edges) 
  (filter 
   (lambda (x) x)
   (map 
    (lambda (edge)
      (let ((nodes (edge-nodes edge)))
        (cond ((equal? (car nodes) node)
               (cadr nodes))
              ((equal? (cadr nodes) node)
               (car nodes))
              (else #f))))
    edges)))

;; Edge operations:
;; Returns a list of all the edges in the graph.
(define get-edges (find-graph-part 'EDGES '()))
;; - checks if two edges are connected.
(define (connected? node1 node2 edges)
  (if (get-edge node1 node2 edges)
      #t
      #f))
;; - gets the distance of an edge.
(define (edge-distance edge)
  (cadr edge))
;; - gets a list of the node id values for the two nodes connected by an edge.
(define (edge-nodes edge)
  (cddr edge))
;; - checks the list of edges and returns an edge if both connected nodes are part of the edge.
(define (get-edge node1 node2 edges)
  (let ((connection
         (filter
          (lambda (edge)
            (if (and (member node1 edge)
                     (member node2 edge))
                #t
                #f))
          edges)))
    (if (> (length connection) 1)
        "Error - more than 1 edge connects two nodes."
        (if (= (length connection) 0) #f
	   (car connection)))))


;; This function takes in a list of node-ids and a list of all nodes in the graph.
;; It returns list of node-ids sorted in order by heuristic value, where the front of the list represents the nodes
;; with the least heuristic value.
(define (sort-by-heuristic node-ids nodes)
  (sort node-ids (lambda (n1 n2) (< (get-heuristic n1 nodes)
				    (get-heuristic n2 nodes)))))

;; Returns a solution for a search.
;; This function was created in order to grade your solutions, so they would have common format.
;; The input extended should be a list of all node ids that you have extended where the node id value of the
;;   first node you extended should be at the front of the list.
;; The input solution-path should be the path found to terminate a search, where it is represented as a list with the
;;   id-value of the start node as the first element in the list.  The last element in the list is the id-node of the
;;   goal node.
(define (create-search-solution extended solution-path)
  (list (cons 'EXTENDED-NODES (list extended))
        (cons 'SOLUTION-PATH (list solution-path))))

;; The graph you will use for the problem set.  The heuristic values are lower bounds on the distance to the node with the id of "CommonArea"
(define thegraph 
  (list '(NODES
          ("Hospital" 17)
          ("Classroom 11" 10)
          ("EntranceHall" 7)
          ("HauntedBathroom" 13)
          ("Dungeon5" 15)
          ("TheChamber" 14)
          ("ForbiddenArea" 8)
          ("TrophyRoom" 6)
          ("Stairs" 4)
          ("GrandHall" 6)
          ("CommonArea" 0)
          ("Statues" 12)
          ("MirroredRoom" 10)
          ("NecessaryRoom" 6))
        '(EDGES 
          (e1 5 "CommonArea" "Stairs")
          (e2 15 "EntranceHall" "Hospital")
          (e3 7 "Classroom 11" "Hospital")
          (e4 25 "HauntedBathroom" "TheChamber")
          (e5 5 "ForbiddenArea" "TrophyRoom")
          (e6 3 "MirroredRoom" "Statues")
          (e7 1 "GrandHall" "EntranceHall")
          (e8 4 "Dungeon5" "HauntedBathroom")
          (e9 2 "Stairs" "GrandHall")
          (e10 9 "Statues" "Stairs")
          (e11 6 "EntranceHall" "HauntedBathroom")
          (e12 4 "ForbiddenArea" "Stairs")
          (e13 10 "Classroom 11" "EntranceHall")
          (e14 5 "TrophyRoom" "Stairs")
          (e15 8 "Stairs" "MirroredRoom")
          (e16 3 "EntranceHall" "Stairs")
          (e17 8 "NecessaryRoom" "CommonArea"))))

;; Another graph you will use for the problem set.  The format of the graph is explained in Problem Set #2 instructions, 
;;   on the class website.
(define graph2
  (list '(NODES
     ("S" 25)
     ("A" 20)
     ("B" 22)
     ("C" 15)
     ("D" 8)
     ("E" 3)
     ("F" 9)
     ("G" 0)) 
   '(EDGES 
     (e1 10 "S" "A")
     (e2 4 "S" "B")
     (e3 9 "A" "C")
     (e4 8 "B" "C")
     (e5 7 "C" "D")
     (e6 9 "C" "E")
     (e7 7 "D" "E")
     (e8 13 "D" "F")
     (e9 8 "E" "F")
     (e10 5 "E" "G")
     (e11 10 "F" "G"))))

;; Another graph you will use for the problem set.  The format of the graph is explained in Problem Set #2 instructions, 
;;   on the class website.
(define graph3
  (list '(NODES
          ("S" 0)
          ("A" 2)
          ("B" 5)
          ("C" 6)
          ("D" 5)
          ("G" 0))
        '(EDGES
          (e1 6 "S" "B")
          (e2 10 "S" "A")
          (e3 10 "A" "B")
          (e4 7 "B" "C")
          (e5 4 "A" "D")
          (e6 2 "C" "D")
          (e7 6 "C" "G")
          (e8 8 "G" "D"))))

;; Another graph you will use for the problem set.  The format of the graph is explained in Problem Set #2 instructions, 
;;   on the class website.
(define graph4
  (list '(NODES
          ("S" 1)
          ("A" 3)
          ("B" 3)
          ("C" 2)
          ("D" 2)
          ("E" 1)
          ("F" 1)
          ("G" 0))
        '(EDGES
          (e1 1 "S" "A")
          (e2 1 "S" "B")
          (e3 1 "A" "B")
          (e4 1 "C" "A")
          (e5 1 "C" "B")
          (e6 1 "D" "C")
          (e7 1 "D" "B")
          (e8 1 "E" "C")
          (e9 1 "E" "D")
          (e10 1 "F" "D")
          (e11 1 "F" "E")
          (e12 1 "G" "E")
          (e13 1 "G" "F"))))

;; Another graph you will use for the problem set.  The format of the graph is explained in Problem Set #2 instructions, 
;;   on the class website.
(define graph5
  (list '(NODES
          ("S" 10)
          ("A" 1000)
          ("B" 5)
          ("C" 5)
          ("G" 0))
        '(EDGES
          (e1 1 "S" "A")
          (e2 1 "G" "C")
          (e3 100 "B" "C")
          (e4 10 "S" "B")
          (e5 10 "C" "A"))))
