(require (lib "list.ss"))
(load "ps4.scm")

;;To grow an identification tree, you need to supply information
;;about the parts of the data you want to classify. In this ID
;;tree system, we accomplish this through *attribute specifications*
;;and *attributes*.
;;
;;Attributes contain enough information to print out a tree-node that
;;splits on the attribute pleasantly, information about how to extract
;;the value of that attribute from a given input, and the range of
;;values that attribute can take on. (The range determines the branches
;;of nodes splitting on that attribute.)
;;
;;Slightly more precisely, an attribute is a list of:
;;- a printable name for the attribute
;;- procedure to extract attribute from an input (feature list)
;;- list of possible values that attribute can take
;;
;;
;;DON'T PANIC. These choices may seem odd, but they give us considerable flexibility
;;in applying our id tree procedure to weird looking data (such as graphs, etc)
;;with complex features (that might need to be computed by a separate, substantial
;;program, such as say connectedness or average path length between two nodes).
;;We won't make you do this, but structuring the code in this way should help
;;shed some light on what information you *actually* need to generate an id tree
;;on potentially arbitrary data.
;;

;;Accessors for an attribute
(define get-attribute-name first)
(define get-attribute-proc second)
(define get-attribute-range third)


;; This procedure makes a classifier from the given attribute specifications
;; (and the error-checking procedure valid-example?)

(define (make-idtree-classifier attribute-specs training-examples)
  (let ([id-tree (make-id-tree attribute-specs training-examples)])
    (lambda (features)
      (run-id-tree id-tree features))))


;; Runs an ID tree on the given feature set to find a classification
(define (run-id-tree id-tree features)
  (if (not (pair? id-tree)) ;;Terminal node
      id-tree
      (let* ([node (first id-tree)]
	     [proc (second node)]
	     [val (proc features)]
	     [range (third node)]
             [children (cdr id-tree)])
	(let child-loop ([range-left range] [kids-left children])
	  (if (or (null? range-left)
		  (null? kids-left))
	      (error "Ran out of tests in ID tree.  Did you construct the tree correctly?\n" id-tree features)
	      (let ([cur-val (car range-left)]
                    [cur-kid (car kids-left)])
		(if (equal? cur-val val)
		    (run-id-tree cur-kid features)
		    (child-loop (cdr range-left) (cdr kids-left)))))))))


;; Returns a pair of the form:
;; ((<class 1> <class 2> .... <class n>) . (<num examples in class 1> ... <num examples
;;   in class n>))
(define (get-classes-with-counts examples)
  ;; Create a hash table to count how many of each class there are,
  ;; using the class as the key and the count as the value
  (let ([classes-counts (make-hash-table 'equal)])
    (let loop ([elts examples])
      (if (not (null? elts))
	  (let* ([cur (car elts)]
		 [class (cdr cur)]
		 [cur-count (hash-table-get classes-counts class #f)])
	    (hash-table-put! classes-counts
			     class
			     (if (eq? #f cur-count)
				 1
				 (+ cur-count 1)))
	    (loop (cdr elts)))))
    
    (let ([classes (hash-table-map classes-counts (lambda (class count) class))]
          [counts (hash-table-map classes-counts (lambda (class count) class))])
      (cons classes counts))))


;; Constructs an ID tree with the given attribute specifications
;; based on the supplied training data.
(define (make-id-tree attribute-specs training-data)
  (let* ([classes-counts (get-classes-with-counts training-data)]
	 [classes (car classes-counts)]
	 [counts (cdr classes-counts)])
    (cond
      [(null? classes) 'Failure]
      [(= (length classes) 1)
       (car classes)]
      [(null? attribute-specs)
       (let* ([classes-counts-combined (map (lambda (class count)
                                              (cons class count))
                                            classes counts)]
              [sorted-classes-counts (sort classes-counts-combined
                                           (lambda (class-count-a class-count-b)
                                             (> (cdr class-count-a) (cdr class-count-b))))])
         (car (first sorted-classes-counts)))]
      [else
       (let* ([attributes-unfiltered attribute-specs]
              [attributes (filter (lambda (x) (not (null? x))) attributes-unfiltered)]
              [disorders-with-attributes
               (map (lambda (attribute)
                      (list (basic-disorder attribute classes training-data)
                            attribute))
                    attributes)]
              [sorted-attributes (sort disorders-with-attributes
                                       (lambda (a b) 
                                         (< (car a) (car b))))]
              [rest-specs (map second (list-tail sorted-attributes 1))]
              [split-attribute-elt (first sorted-attributes)]
              [split-attribute (second split-attribute-elt)]
              [split-attribute-range (third split-attribute)]
              [split-attribute-proc (second split-attribute)]
              [split-attribute-name (first split-attribute)]
              [split-data (map (lambda (attribute-val)
                                 (filter (lambda (example)
                                           (equal? (split-attribute-proc (get-features example)) attribute-val))
                                         training-data))
                               split-attribute-range)])
         (cons (list split-attribute-name
                     split-attribute-proc
                     split-attribute-range)
               (map (lambda (data-for-branch)
                      (make-id-tree rest-specs data-for-branch))
                    split-data)))])))
